/*
 * Copyright (C) 1992, 1993, 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */
/***************************************************************************
 *
 * @(#) - BZ - Multiplayer tank game - Host validation routines.
 *
 *			        Chris Fouts - Silicon Graphics, Inc.
 *			        October, 1991
 **************************************************************************/
#include <stdio.h>
#include <stdarg.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <time.h>

#include <X11/Intrinsic.h>

#include "bz.h"
#include "bzvalidate.h"
#include "bzbase.h"
#include "framerate.h"

#define	MAX_DROPS	2
#define	MAX_QUITS	3
#define	MAX_ACKS	20

#define	BZ_STEADY_CENSUS_RATE	400			/* tenths o' seconds */
#define	BZ_UNSTEADY_CENSUS_RATE	 40			/* tenths o' seconds */
#define BZ_NEW_DEADLINE_TIME	 50			/* seconds */
#define BZ_OLD_DEADLINE_TIME	120			/* seconds */
#define	BZ_WEEDING_TIME			 15			/* seconds */
#define	BZ_DROPPING_DELTA		-30			/* seconds */

#define BZ_VALID_UNKNOWN		0
#define BZ_VALID				1
#define BZ_INVALID				2
#define BZ_DROPPED				3
#define BZ_VALID_QUIT			4

struct bz_host {
	long			id ;
	short			agrees ;
	short			drops ;
	short			quits ;
	short			acks ;
	short			valid ;
	short			last_valid ;
	struct bz_host	*next ;
	} ;

struct bz_census {
	struct BzCensus		census ;
	struct bz_census	*next ;
	} ;

/* BEGIN PROTOTYPES -S bzvalidate.c */
static struct bz_host  *add_host( long host_id ) ;
static int              compare_census( struct BzCensus *input ) ;
static void             copybytes( void *src, void *dst, long n ) ;
static void             do_weeding( void ) ;
static struct bz_host  *find_host( long host_id ) ;
static int              find_host_in_my_census_list( long id ) ;
static void             generate_census_checksum( void ) ;
static void             list_hosts( void ) ;
static void             reset_deadlines( int except ) ;
static void             show_host_validity( struct bz_host *host ) ;
/* END PROTOTYPES -S bzvalidate.c */

static struct bz_host	*first_host = NULL ;
static unsigned short	game_key = 0 ;
static struct timeval	last_census_sent ;
static long				time_between_census_posts = 0 ;
static Boolean			need_to_send_census ;
static Boolean			need_to_do_weeding = FALSE ;
extern struct timeval	current_time ;
extern struct timeval	start_up_time ;
static struct timeval	weed_time ;
static struct bz_census	my_census ;
extern struct BzMember	player[] ;
extern struct timeval	deadline[] ;
extern unsigned int		number_players ;



/*------------------------------------------------------------------------------
 * Find a host within the host list.
 *----------------------------------------------------------------------------*/
static struct bz_host *find_host(
	long host_id
	)
{
	struct bz_host	*host = first_host ;

	while( host != NULL ) {
		if( host->id == host_id ) {
			return( host ) ;
			}
		else {
			host = host->next ;
			}
		}

	return( host ) ;
}



/*------------------------------------------------------------------------------
 * Add a host to the host list.
 *----------------------------------------------------------------------------*/
static struct bz_host *add_host(
	long host_id
	)
{
	struct bz_host	*host = first_host ;
	struct bz_host	*prev = NULL ;

	while( host != NULL ) {
		prev = host ;
		host = host->next ;
		}

	if( ( host = (struct bz_host *)malloc( sizeof( struct bz_host ) ) ) !=
		NULL ) {
		host->id = host_id ;
		host->agrees = 0 ;
		host->drops = 0 ;
		host->quits = 0 ;
		host->acks = 0 ;
		host->valid = BZ_VALID_UNKNOWN ;
		host->last_valid = BZ_VALID_UNKNOWN ;
		host->next = NULL ;
		if( first_host == NULL ) {
			first_host = host ;
			}
		dprintf( "%s: started receiving from %s\n",
				  print_time(), host_name_from_id( host_id ) ) ;
		}

	if( prev != NULL ) {
		prev->next = host ;
		}

	return( host ) ;
}



/*------------------------------------------------------------------------------
 * Remove a host from the host list.
 *----------------------------------------------------------------------------*/
void remove_host(
	long host_id
	)
{
	struct bz_host	*host = first_host ;
	struct bz_host	*prev = NULL ;

	while( host != NULL && host->id != host_id ) {
		prev = host ;
		host = host->next ;
		}

	if( host == NULL )
		return ;

	if( host->id == host_id ) {
		/*
		 * If prev == NULL, this means we're removing the first host.
		 * Set the first_host to point to the next host.
		 */
		if( prev == NULL ) {
			first_host = host->next ;
			}
		/*
		 * We're past the first host, so link up the previous one to
		 * the next one.
		 */
		else {
			prev->next = host->next ;
			}
		free( host ) ;
		}

#if defined( NETDEBUGGER )
	list_hosts() ;
#endif /* defined( NETDEBUGGER ) */

	return ;
}



/*------------------------------------------------------------------------------
 * Returns non-zero if the maximum number of acknowledgement packets have
 * not been sent.
 *----------------------------------------------------------------------------*/
int should_acknowledge(
	long host_id,
	unsigned short key
	)
{
	struct bz_host	*host ;
	int				st ;

	if( key != game_key ) {
		return( 0 ) ;
		}

	if( ( host = find_host( host_id ) ) == NULL ) {
		if( ( host = add_host( host_id ) ) == NULL ) {
			perror( "should_acknowledge" ) ;
			end_program( 1 ) ;
			}
		}

	if( host->acks < MAX_ACKS && host->valid == BZ_VALID_UNKNOWN ) {
		host->acks++ ;
		st = 1 ;
		if( host->acks == 1 ) {
			dprintf( "%s: should acknowledge %s.\n", print_time(),
				host_name_from_id( host_id ) );
			}
		}
	else {
		st = 0 ;
		}


	return( st ) ;
}



/*------------------------------------------------------------------------------
 * Called *only* when an ACKNOWLEDGEMENT packet is received (or in view-only
 * mode).
 *---------------------------------------------------------------------------*/
int valid_keyed_host(
	long host_id,
	unsigned short key
	)
{
	struct bz_host	*host ;

	if( key != game_key ) {
		return( 0 ) ;
		}

	if( ( host = find_host( host_id ) ) == NULL ) {
		if( ( host = add_host( host_id ) ) == NULL ) {
			perror( "valid_keyed_host" ) ;
			end_program( 1 ) ;
			}
		}

	if( host->valid == BZ_VALID_UNKNOWN ) {
		host->valid = BZ_VALID ;
		dprintf( "%s: acknowledged by %s\n",
				  print_time(), host_name_from_id( host_id ) ) ;
		}
	/*
	 * If host had dropped me, it's now receiving my packets, so send an
	 * ACK packet so it will know I see it and add me to its game.
	 */
	else if( host->valid == BZ_DROPPED ) {
		host->valid = BZ_VALID ;
		send_acknowledgement( host_id ) ;
		dprintf( "%s: re-acknowledged by %s (had dropped me)\n",
				  print_time(), host_name_from_id( host_id ) ) ;
		}

	return( host->valid == BZ_VALID ) ;
}



/*------------------------------------------------------------------------------
 * Returns TRUE if a host is valid.
 *----------------------------------------------------------------------------*/
int valid_host(
	long host_id
	)
{
	struct bz_host	*host ;

	if( ( host = find_host( host_id ) ) == NULL ) {
		return( 0 ) ;
		}

	/*
	 * If host had quit, he has started back up again, so reset its status.
	 */
	if( host->valid == BZ_VALID_QUIT ) {
		host->valid = BZ_VALID_UNKNOWN ;
		dprintf( "%s: receiving from %s again (had quit)\n",
				  print_time(), host_name_from_id( host_id ) ) ;
		}

	return( host->valid == BZ_VALID ) ;
}



/*------------------------------------------------------------------------------
 * Note that a invalid host quit.
 *---------------------------------------------------------------------------*/
void quit_host(
	long host_id
	)
{
	struct bz_host	*host ;
	struct BzAck			invalid ;

	if( ( host = find_host( host_id ) ) == NULL ) {
		return ;
		}

	if( host->valid != BZ_VALID_QUIT && host->quits < MAX_QUITS ) {
		host->quits += 1 ;
		host->last_valid = host->valid ;
		host->valid = BZ_VALID_QUIT ;
		host->acks = 0 ;
		dprintf( "%s: %s quit\n", print_time(), host_name_from_id( host_id ) ) ;
		}
}



/*------------------------------------------------------------------------------
 * Drop a host when I haven't received any packets in a while from it.
 *---------------------------------------------------------------------------*/
void drop_host(
	long host_id
	)
{
	struct bz_host	*host ;
	struct BzAck			invalid ;

	if( ( host = find_host( host_id ) ) == NULL ) {
		return ;
		}
	host->drops += 1 ;

	if( host->drops > MAX_DROPS ) {
		invalidate_host( host_id ) ;
		}
	else {
		/*
		 * Notify host of being dropped.
		 */
		invalid.bz_id = BZDROP ;
		invalid.id = player[SELF].id ;
		invalid.ack_id = host_id ;
		invalid.key = game_key ;
		send_out( &invalid, sizeof( invalid ) ) ;
#if defined( NETDEBUGGER )
		post_new_message( 0, 0, "sent drop packet to %c", 'a' + host_id ) ;
#endif /* defined( NETDEBUGGER ) */

		/*
		 * Reset the host so that when I start receiving packets from it
		 * again, I'll acknowledge it.
		 */
		host->valid = BZ_VALID_UNKNOWN ;
		host->acks = 0 ;
		dprintf( "%s: dropping %s from game (stopped getting packets)\n",
				  print_time(), host_name_from_id( host_id ) ) ;
		}
}



/*------------------------------------------------------------------------------
 * I've been dropped by a host, so indicate status with that host so that I
 * can know when it starts receiving me again.
 *---------------------------------------------------------------------------*/
void dropped_by_host(
	long host_id
	)
{
	struct bz_host			*host ;

	if( ( host = find_host( host_id ) ) != NULL ) {
		host->valid = BZ_DROPPED ;
		dprintf( "%s: removing %s from game (dropped by him)\n",
				  print_time(), host_name_from_id( host_id ) ) ;
		}
}



/*------------------------------------------------------------------------------
 * Invalidate a host.  This is permanent.  Will ignore all future packets
 * from this host (until it quits).
 *---------------------------------------------------------------------------*/
void invalidate_host(
	long host_id
	)
{
	int						i ;
	int						j ;
	struct bz_host			*host ;
	struct BzAck			invalid ;
	struct bz_census		*cen ;

	if( ( host = find_host( host_id ) ) != NULL ) {
		host->valid = BZ_INVALID ;

		/*
		 * Notify host of being invalidated.
		 */
		invalid.bz_id = BZINVALIDATE ;
		invalid.id = player[SELF].id ;
		invalid.ack_id = host_id ;
		invalid.key = game_key ;
		send_out( &invalid, sizeof( invalid ) ) ;
		if( ( cen = find_host_census( host_id ) ) != NULL ) {
			dprintf( "%s: %s invalidated --\n",
				     print_time(), host_name_from_id( host_id ) ) ;
			dprintf( "\tHe sees %ld hosts while you see %ld.\n",
					 cen->census.n_players, my_census.census.n_players ) ;
			dprintf( "\tThe following hosts were seen by him:\n" ) ;
			for( i = 0 ; i < cen->census.n_players ; i++ ) {
				if( cen->census.list[i] == player[SELF].id ) {
					dprintf( "\t\t%s (valid - self)\n",
							host_name_from_id( cen->census.list[i] ) ) ;
					}
				else {
					host = find_host( cen->census.list[i] ) ;
					if( host == NULL ) {
						dprintf( "\t\t%s (do not see)\n",
								host_name_from_id( cen->census.list[i] ) ) ;
						}
					else {
						show_host_validity( host ) ;
						}
					}
				}
			dprintf( "\tThe following hosts were seen by you but "
								"not validated by him:\n" ) ;

			j = 0 ;
			for( i = 0 ; i < my_census.census.n_players ; i++ ) {
				while( j < cen->census.n_players &&
					   my_census.census.list[i] > cen->census.list[j] ) {
					j++ ;
					}
				if( j < cen->census.n_players &&
					my_census.census.list[i] == cen->census.list[j] ) {
					j++ ;
					}
				else if( my_census.census.list[i] != player[SELF].id ) {
					dprintf( "\t\t%s\n",
							host_name_from_id( my_census.census.list[i] ) ) ;
					}
				}
			dprintf( "end list for %s\n\n", host_name_from_id( host_id ) ) ;
			}
		}
}



/*------------------------------------------------------------------------------
 * Print out the valid of a host.
 *----------------------------------------------------------------------------*/
static void show_host_validity(
	struct bz_host *host
	)
{

	if( host->valid == BZ_INVALID ) {
		dprintf( "\t\t%s (invalidated)\n",
				host_name_from_id( host->id ) ) ;
		}
	else if( host->valid == BZ_VALID_UNKNOWN ) {
		if( host->drops ) {
			dprintf( "\t\t%s (lost communication from)\n",
					host_name_from_id( host->id ) ) ;
			}
		else {
			dprintf( "\t\t%s (does not see me)\n",
					host_name_from_id( host->id ) ) ;
			}
		}
	else if( host->valid == BZ_VALID_QUIT ) {
		if( host->last_valid == BZ_VALID_QUIT ) {
			dprintf( "\t\t%s (thought had quit)\n",
					host_name_from_id( host->id ) ) ;
			}
		else {
			host->valid = host->last_valid ;
			show_host_validity( host ) ;
			host->valid = BZ_VALID_QUIT ;
			}
		}
	else {
		dprintf( "\t\t%s (valid)\n",
				host_name_from_id( host->id ) ) ;
		}
}



/*------------------------------------------------------------------------------
 * Register the game's key code.
 *----------------------------------------------------------------------------*/
void register_key(
	unsigned short key
	)
{
	game_key = key ;
}



/*------------------------------------------------------------------------------
 * Generate the census checksum (both xor and summation) from the list of
 * hostid's in the census host list.
 *----------------------------------------------------------------------------*/
static void generate_census_checksum( void )
{
	int	i ;

	my_census.census.chksum = 0 ;
	my_census.census.chkxor = 0 ;
	for( i = 0 ; i < my_census.census.n_players ; i++ ) {
		my_census.census.chksum = ( my_census.census.chksum +
					( my_census.census.list[i] & 0x3fffffff ) ) & 0x3fffffff ;
		my_census.census.chkxor ^= my_census.census.list[i] ;
		}
}



/*------------------------------------------------------------------------------
 * Initialize the internal census packet.
 *----------------------------------------------------------------------------*/
void init_census_packet(
	long id
	)
{
	my_census.census.bz_id     = BZCENSUS ;
	my_census.census.id        = id ;
	my_census.census.n_players = 1 ;
	my_census.census.list[0]   = id ;
	my_census.next			   = NULL ;
	copy_time( &last_census_sent, &current_time ) ;

	set_time( &weed_time, &current_time, BZ_WEEDING_TIME, 0 ) ;
	need_to_do_weeding = FALSE ;

	generate_census_checksum() ;
}



/*------------------------------------------------------------------------------
 * Move bytes from one source to another, allowing for overlap
 *----------------------------------------------------------------------------*/
static void copybytes(
	void *src,
	void *dst,
	long n
	)
{
	char	*s ;
	char	*d ;

	if( src > dst ) {
		s = (char *)src ;
		d = (char *)dst ;
		while( n-- )
			*(d++) = *(s++) ;
		}
	else {
		s = (char *)src + n ;
		d = (char *)dst + n ;
		while( n-- )
			*(--d) = *(--s) ;
		}
}



/*------------------------------------------------------------------------------
 * Add a host to our census list.
 *----------------------------------------------------------------------------*/
void add_host_to_census(
	long id
	)
{
	int						n ;
	int						i ;
	int						insert_pt ;
	static int				first = 1 ;
	struct bz_census		*next = my_census.next ;
	struct bz_census		*prev = &my_census ;

	if( find_host_in_my_census_list( id ) != -1 ) {
		return ;
		}

	if( ( i = find_player( id ) ) == -1 ) {
		return ;
		}

	n = (my_census.census.n_players)++ ;

	while( next ) {
		prev = next ;
		next = next->next ;
		}

	if( ( next = calloc( 1, sizeof( struct bz_census ) ) ) == NULL ) {
		perror( "add_host_to_census" ) ;
		end_program( 1 ) ;
		}

	prev->next = next ;

	next->census.id = id ;
	next->census.n_players = 1 ;
	next->census.chksum    = id ;
	next->census.chkxor    = id ;

	insert_pt = n ;
	while( insert_pt && my_census.census.list[insert_pt-1] > id )
		insert_pt-- ;

	copybytes( &(my_census.census.list[insert_pt]),
			   &(my_census.census.list[insert_pt+1]),
			   ( n - insert_pt ) * sizeof( my_census.census.list[0] ) ) ;

	my_census.census.list[insert_pt] = id ;

	generate_census_checksum() ;

	if( !first ) {
		reset_deadlines( i ) ;
		}
	first = 0 ;

	set_time( &weed_time, &current_time, BZ_WEEDING_TIME, 0 ) ;
	need_to_do_weeding = TRUE ;

	set_time( &(deadline[i]), &current_time, BZ_NEW_DEADLINE_TIME, 0 ) ;
	dprintf( "%s: added %s to game and census\n", print_time(),
			host_name_from_id( id ) ) ;

#if defined( NETDEBUGGER )
	post_new_message( 0, 0, "added host %s to census, deadline = %s",
						player[i].name,
						print_time_hundreths( &(deadline[i]) ) ) ;
#endif /* defined( NETDEBUGGER ) */

	need_to_send_census = TRUE ;
}



/*------------------------------------------------------------------------------
 * Find a host in the linked list of census packets.
 *----------------------------------------------------------------------------*/
struct bz_census *find_host_census(
	long id
	)
{
	struct bz_census		*next = my_census.next ;

	while( next && next->census.id != id ) {
		next = next->next ;
		}

	return( next ) ;
}



/*------------------------------------------------------------------------------
 * Remove a host from our census list.
 *----------------------------------------------------------------------------*/
void remove_host_from_census(
	long id
	)
{
	int	n ;
	struct bz_census		*next = my_census.next ;
	struct bz_census		*prev = &my_census ;

	n = find_host_in_my_census_list( id ) ;

	while( next && next->census.id != id ) {
		prev = next ;
		next = next->next ;
		}

	if( n < 0 || next == NULL ) {
		return ;
		}

	prev->next = next->next ;

	free( next ) ;

	my_census.census.n_players-- ;

	copybytes( &(my_census.census.list[n+1]), &(my_census.census.list[n]),
		(my_census.census.n_players - n) * sizeof( my_census.census.list[0] ) );

	generate_census_checksum() ;

	need_to_send_census = TRUE ;
}



/*------------------------------------------------------------------------------
 * Locate a host within our census list.
 *----------------------------------------------------------------------------*/
static int find_host_in_my_census_list(
	long id
	)
{
	int		n ;
	int		lo = 0 ;
	int		hi = my_census.census.n_players - 1 ;
	long	*census_list = my_census.census.list ;

	while( 1 ) {
		n = ( lo + hi ) / 2 ;
		if( id == census_list[n] ) {
			return( n ) ;
			}
		else if( id < census_list[n] ) {
			hi = n - 1 ;
			}
		else {
			lo = n + 1 ;
			}
		if( lo > hi ) {
			return( -1 ) ;
			}
		}
}



/*------------------------------------------------------------------------------
 * Compare a census packet with another.
 *----------------------------------------------------------------------------*/
static int compare_census(
	struct BzCensus *input
	)
{
	return( ( input->n_players == my_census.census.n_players ) &&
			( input->chksum == my_census.census.chksum ) &&
			( input->chkxor == my_census.census.chkxor ) ) ;
}



/*------------------------------------------------------------------------------
 * Process a census packet.
 *----------------------------------------------------------------------------*/
void process_bz_census(
	struct BzCensus	*input
	)
{
	int						n ;
	int						i ;
	struct bz_census		*pcen ;
	struct bz_host			*host ;

	if( ( n = find_player( input->id ) ) > 0 ) {
		/*
		 * Save the census packet.
		 */
		if( ( pcen = find_host_census( input->id ) ) == NULL ) {
#if defined( NETDEBUGGER )
			post_new_message( 0, 0, "could not find CENSUS for %c",
							  'a' + input->id ) ;
#endif /* defined( NETDEBUGGER ) */
			return ;
			}
		else {
			copybytes( input, &(pcen->census), sizeof( struct BzCensus ) ) ;
			}
		/*
		 * Check if census matches up with SELF's census.
		 */
		if( compare_census( input ) ) {
			/*
			 * It matches, so set new deadline for player.
			 */
			set_time( &(deadline[n]), &current_time, BZ_OLD_DEADLINE_TIME, 0 ) ;
			if( ( host = find_host( input->id ) ) != NULL ) {
				if( host->agrees == 0 ) {
					dprintf( "%s: validation now matching from %s.\n",
							 print_time(), host_name_from_id( host->id ) ) ;
					host->agrees = 1 ;
					}
				}
			}
		else {
			/*
			 * No match.  See if player's deadline has been exceeded.
			 */
			if( ( host = find_host( input->id ) ) != NULL ) {
				if( host->agrees == 1 ) {
					dprintf( "%s: validation no longer matches from %s.\n",
							 print_time(), host_name_from_id( host->id ) ) ;
					host->agrees = 0 ;
					}
				}
			if( elapsed_sec_tenths( &(deadline[n]), &current_time ) > 0 ) {
				/*
				 * Deadline passed, drop player.
				 */
				invalidate_host( input->id ) ;
				delete_player( n, "was invalidated" ) ;
				}

			/*
			 * Set time between census broadcasts to a smaller number since
			 * the census lists are in flux.
			 */
			time_between_census_posts = BZ_UNSTEADY_CENSUS_RATE ;
			}
		}
}



/*------------------------------------------------------------------------------
 * Check to see if we need to send a census packet.
 *----------------------------------------------------------------------------*/
void check_to_send_census( void )
{
	long	tenths ;

	if( need_to_do_weeding &&
		elapsed_sec_tenths( &weed_time, &current_time ) > 0 ){
		do_weeding() ;
		need_to_do_weeding = FALSE ;
		}

	if( need_to_send_census ||
		elapsed_sec_tenths( &last_census_sent, &current_time ) >
					time_between_census_posts ) {
		need_to_send_census = FALSE ;
		send_out( &(my_census.census), sizeof( struct BzCensus ) ) ;
		copy_time( &last_census_sent, &current_time ) ;
		time_between_census_posts = BZ_STEADY_CENSUS_RATE ;
		}
}



/*------------------------------------------------------------------------------
 * Reset all of the census deadlines except for the newest.
 *----------------------------------------------------------------------------*/
static void reset_deadlines(
	int except
	)
{
	int						i ;
	long					secs ;
	long					tenths ;

	if( ( secs = elapsed_sec_tenths( &last_census_sent, &current_time ) ) <
			BZ_OLD_DEADLINE_TIME * 10 ) {

		tenths = secs % 10 ;
		secs /= 10 ;
#if defined( NETDEBUGGER )
		post_new_message( 0, 0, "%s increasing deadline by %ld.%ld secs",
						  print_time(), secs, tenths ) ;
#endif /* defined( NETDEBUGGER ) */

		for( i = ENEMY ; i < number_players ; i++ ) {
			if( i != except ) {
				set_time( &(deadline[i]), &(deadline[i]), secs, tenths ) ;
#if defined( NETDEBUGGER )
				dprintf( "deadline for %s is %s",
						host_name_from_id( player[i].id ),
						print_time_hundreths( &(deadline[i]) ) ) ;
#endif /* defined( NETDEBUGGER ) */
				}
			}
		}
}



/*------------------------------------------------------------------------------
 * Print out a list of all non-valid hosts and a guess at why.
 *----------------------------------------------------------------------------*/
void validation_summary( void )
{
	struct bz_host	*host = first_host ;

	while( host ) {
		switch( host->valid ) {

			case BZ_VALID_UNKNOWN :
				if( host->drops ) {
					dprintf( "Lost communcation from host %s.\n",
							 host_name_from_id( host->id ) ) ;
					}
				else {
					dprintf( "Host %s evidently could not receive my packets."
							 "\n", host_name_from_id( host->id ) ) ;
					}
				break ;

			case BZ_VALID :
				dprintf( "Host %s was valid.\n",
						 host_name_from_id( host->id ) ) ;
				break ;

			case BZ_INVALID :
				break ;

			case BZ_VALID_QUIT :
				dprintf( "Host %s has quit (validity unknown).\n",
						 host_name_from_id( host->id ) ) ;
				break ;

			case BZ_DROPPED :
				dprintf(
						 "Host %s evidently stopped receiving my packets.\n",
						 host_name_from_id( host->id ) ) ;
				break ;

			default :
				dprintf( "Odd validity for %s.\n",
						 host_name_from_id( host->id ) ) ;
				break ;
			}
		host = host->next ;
		}

	dprintf( "end" ) ;
}



/*------------------------------------------------------------------------------
 * Convert a `hostid' number to a string representing the host's name if
 * known, otherwise the IP address.
 *----------------------------------------------------------------------------*/
char *host_name_from_id(
	long id
	)
{
	struct in_addr	addr ;
	struct hostent	*hp ;
	static char		buffer[256] ;

#if !defined( NETDEBUGGER )
	addr.s_addr = id ;
	hp = gethostbyaddr( &addr, sizeof(addr), AF_INET ) ;
	if( hp ) {
		sprintf( buffer, "%s (%s)", hp->h_name, inet_ntoa(addr) ) ;
		}
	else {
		sprintf( buffer, "%s", inet_ntoa(addr) ) ;
		}
#else
	sprintf( buffer, "player %c", 'a' + id ) ;
#endif /* !defined( NETDEBUGGER ) */

	return( buffer ) ;
}



/*------------------------------------------------------------------------------
 * Send out an ACK packet.
 *---------------------------------------------------------------------------*/
void send_acknowledgement(
	long id
	)
{
	struct BzAck			ack ;

	ack.bz_id = BZACK ;
	ack.id = player[SELF].id ;
	ack.ack_id = id ;
	ack.key = player[SELF].key ;
	send_out( &ack, sizeof( ack ) ) ;
}



/*------------------------------------------------------------------------------
 * Spit out a list of all the hosts.
 *----------------------------------------------------------------------------*/
static void list_hosts( void )
{
	int				i = 0 ;
	struct bz_host	*host = first_host ;

	while( host != NULL ) {
#if defined( NETDEBUGGER )
		post_new_message( 0, 0, "%d - %c (%hd,%hd) %hd", i, 'a'+host->id,
						  host->drops, host->acks, host->valid ) ;
#else
		post_new_message( 0, 0, "%d - %s (%hd,%hd) %hd", i,
						  host_name_from_id( host->id ), host->drops,
						  host->acks, host->valid ) ;
#endif /* defined( NETDEBUGGER ) */
		host = host->next ;
		}

	return ;
}



/*------------------------------------------------------------------------------
 * Do initial weeding of host to try and join the largest group.
 *----------------------------------------------------------------------------*/
static void do_weeding( void )
{
	int						i ;
	int						j ;
	int						n ;
	int						total ;
	int						best_total = 0 ;
	int						least_seen ;
	int						times_seen[MAXPLAYERS] ;
	struct bz_census		*next = my_census.next ;
	struct bz_census		*prev = &my_census ;
	struct bz_census		*best ;

#if defined( NETDEBUGGER )
	post_new_message( 0, 0, "%s doing initial weeding", print_time() ) ;
#endif /* defined( NETDEBUGGER ) */

	/*
	 * Go through all players in my census and mark as being seen once (by me).
	 */
	for( i = 0 ; i < my_census.census.n_players ; i++ ) {
		times_seen[i] = 1 ;
		}

	/*
	 * Go through all other players and mark players according to how many
	 * times they appear on other's censuses.
	 */
	while( next ) {
		total = 0 ;
		for( i = 0 ; i < next->census.n_players ; i++ ) {
			/*
			 * Check if host is on my census.
			 */
			n = find_host_in_my_census_list( next->census.list[i] ) ;
			/*
			 * If host is in my census, increment total and mark player.
			 */
			if( n != -1 ) {
				total++ ;
				times_seen[n] += 1 ;
				}
			}
		/*
		 * Check number of census matches.
		 */
		if( total > best_total ) {
			best = next ;
			best_total = total ;
			}
		next = next->next ;
		}

#if defined( NETDEBUGGER )
	for( i = 0 ; i < my_census.census.n_players ; i++ ) {
		post_new_message( 0, 0, "*        * %c seen %d times",
			'a' + my_census.census.list[i], times_seen[i] ) ;
		}
#endif /* defined( NETDEBUGGER ) */

	/*
	 * If the best match is greater than zero...
	 */
	if( best_total > 0 && best != NULL ) {

#if defined( NETDEBUGGER )
		post_new_message( 0, 0, "*        * best total = %d/%d (player %c)",
							best_total, best->census.n_players,
							'a' + best->census.id ) ;
#endif /* defined( NETDEBUGGER ) */

		/*
		 * Compare my census with the best match.
		 */
		j = 0 ;
		least_seen = best_total + 1 ;
		for( i = 0 ; i < my_census.census.n_players ; i++ ) {
			/*
			 * Find location in best census where my i-th census entry
			 * should be.
			 */
			while( j < best->census.n_players &&
				   my_census.census.list[i] > best->census.list[j] ) {
				j++ ;
				}
			/*
			 * If i-th entry is in best census, check for least seen value.
			 */
			if( j < best->census.n_players &&
				my_census.census.list[i] == best->census.list[j] ) {
				if( times_seen[i] <= least_seen ) {
					least_seen = times_seen[i] ;
					}
				j++ ;
				}
			/*
			 * The i-th entry isn't in the best census, so shorten it's
			 * deadline.
			 */
			else {
				if( ( n = find_player( my_census.census.list[i] ) ) > 0 ) {
					set_time( &(deadline[n]), &(deadline[n]),
							  BZ_DROPPING_DELTA, 0 ) ;
#if defined( NETDEBUGGER )
					post_new_message( 0, 0, "*        * should_drop %c (%s)",
						'a' + my_census.census.list[i],
						print_time_hundreths( &(deadline[n]) ) ) ;
#endif /* defined( NETDEBUGGER ) */
					}
				}
			}

#if defined( NETDEBUGGER )
		post_new_message( 0, 0, "*        * least_seen = %d", least_seen ) ;
#endif /* defined( NETDEBUGGER ) */

		/*
		 * If some entry is seen less times than the best total, shorten
		 * its deadline time.
		 */
		if( least_seen < best_total ) {
			for( i = 0 ; i < my_census.census.n_players ; i++ ) {
				if( times_seen[i] == least_seen ) {
					/*
					 * Decrease deadline.
					 */
					if( ( n = find_player( my_census.census.list[i] ) ) > 0 ) {
						set_time( &(deadline[n]), &(deadline[n]),
								  BZ_DROPPING_DELTA, 0 ) ;
#if defined( NETDEBUGGER )
						post_new_message( 0, 0,
							"*        * should_drop %c (%s)",
							'a' + my_census.census.list[i],
							print_time_hundreths( &(deadline[n]) ) ) ;
#endif /* defined( NETDEBUGGER ) */
						}
					}
				}
			}
		}

#if defined( NETDEBUGGER )
	post_new_message( 0, 0, "********** ending weeding" ) ;
#endif /* defined( NETDEBUGGER ) */
}



/*------------------------------------------------------------------------------
 * Debug printf.
 *----------------------------------------------------------------------------*/
void dprintf(
	char *fmt,
	...
	)
{
	va_list		args ;
	static int	status = 0 ;
	static FILE	*debug_out ;

	if( status == -1 ) {
		return ;
		}

	if( strcmp( fmt, "end" ) == 0 ) {
#if !defined( NETDEBUGGER )
		if( status == 1 ) {
			fclose( debug_out ) ;
			printf( "Summary of validations written to /usr/tmp/bz.log\n" ) ;
			}
#endif /* !defined( NETDEBUGGER ) */
		return ;
		}
	else if( strcmp( fmt, "disable" ) == 0 ) {
		status = -1 ;
		return ;
		}

	if( status == 0 ) {
#if defined( NETDEBUGGER )
		status = 1 ;
#else
		if( ( debug_out = fopen( "/usr/tmp/bz.log", "w" ) ) == NULL ) {
			status = -1 ;
			return ;
			}
		else {
			status = 1 ;
			}
#endif /* defined( NETDEBUGGER ) */
		}

	va_start( args, fmt ) ;
#if defined( NETDEBUGGER )
	print_to_net_debugger( fmt, args ) ;
#else
	vfprintf( debug_out, fmt, args ) ;
#endif /* defined( NETDEBUGGER ) */
	va_end( args ) ;
}



/*------------------------------------------------------------------------------
 * Print out the time.
 *----------------------------------------------------------------------------*/
char *print_time( void )
{
	static char	timestr[64] ;
	time_t		clock_val ;

	clock_val = time( NULL ) ;

	cftime( timestr, "%r", &clock_val ) ;

	return( timestr ) ;
}



/*------------------------------------------------------------------------------
 * Print out the date.
 *----------------------------------------------------------------------------*/
char *print_date( void )
{
	static char	datestr[64] ;
	time_t		clock_val ;

	clock_val = time( NULL ) ;

	cftime( datestr, "%D", &clock_val ) ;

	return( datestr ) ;
}



/*------------------------------------------------------------------------------
 * Print out the time with hundreths of seconds.
 *----------------------------------------------------------------------------*/
char *print_time_hundreths(
	struct timeval *t
	)
{
	static char	timestr[64] ;
	int			i ;
	time_t		secs ;
	long		hsecs ;

	secs = t->tv_sec + t->tv_usec / 1000000 ;
	hsecs = ( t->tv_usec % 1000000 ) / 10000 ;
	cftime( timestr, "%I:%M:%S.00 %p", &secs ) ;
	i = strlen( timestr ) ;

	sprintf( timestr + 9, "%02d", hsecs ) ;
	timestr[11] = ' ' ;

	return( timestr ) ;
}



